from dataclasses import dataclass

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.text.bert import _model as bert

__all__ = ["RoFormerConfig", "RoFormer"]


@dataclass
class RoFormerConfig(bert.BERTConfig):
    rotary_value: bool = False
    rope_interleaved: bool = True


class _RoFormerEmbeddings(nn.Module):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__()
        self.word_embeddings = nn.Embedding(
            config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id
        )
        self.token_type_embeddings = nn.Embedding(
            config.type_vocab_size, config.hidden_size
        )

        self.layernorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

        self.token_type_ids: nn.Buffer
        self.register_buffer(
            "token_type_ids",
            nn.Buffer(lucid.zeros(1, config.max_position_embeddings), dtype=lucid.Long),
        )

    def forward(
        self,
        input_ids: lucid.LongTensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
        past_key_values_length: int = 0,
        cache_position: Tensor | None = None,
    ) -> Tensor:
        del position_ids, past_key_values_length, cache_position

        if input_ids is not None:
            input_shape = input_ids.shape
            device = input_ids.device
        else:
            input_shape = inputs_embeds.shape[:-1]
            device = inputs_embeds.device

        batch_size, seq_length = input_shape
        if token_type_ids is None:
            if self.token_type_ids.shape[1] >= seq_length:
                token_type_ids = self.token_type_ids[:, :seq_length].expand(
                    batch_size, seq_length
                )
            else:
                token_type_ids = lucid.zeros(
                    *input_shape, dtype=lucid.Long, device=device
                )

        elif token_type_ids.shape != input_shape:
            raise ValueError(
                "'token_type_ids' must match input token shape "
                f"(expected {input_shape}, got {token_type_ids.shape})"
            )

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)

        token_type_embeddings = self.token_type_embeddings(token_type_ids)
        embeddings = inputs_embeds + token_type_embeddings

        embeddings = self.layernorm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class _RoFormerSelfAttention(bert._BERTSelfAttention):
    def __init__(self, config: RoFormerConfig, /, layer_idx: int | None = None) -> None:
        super().__init__(config, layer_idx)
        self.rotary_value = config.rotary_value

        self.rope_qk = nn.RotaryPosEmbedding(
            embed_dim=self.attention_head_size,
            max_seq_len=config.max_position_embeddings,
            interleaved=config.rope_interleaved,
        )
        self.rope_v = (
            nn.RotaryPosEmbedding(
                embed_dim=self.attention_head_size,
                max_seq_len=config.max_position_embeddings,
                interleaved=config.rope_interleaved,
            )
            if config.rotary_value
            else None
        )

    def _resolve_rope_position_ids(
        self,
        *,
        hidden_states: Tensor,
        position_ids: Tensor | None,
        cache_position: Tensor | None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None,
    ) -> Tensor:
        batch_size, seq_length = hidden_states.shape[:2]
        device = hidden_states.device

        if position_ids is not None:
            if position_ids.ndim == 1:
                if position_ids.shape[0] != seq_length:
                    raise ValueError(
                        "1-D position_ids must have length == seq_length "
                        f"(got {position_ids.shape[0]} vs {seq_length})."
                    )
                return position_ids.to(device).long()

            if (
                position_ids.ndim != 2
                or position_ids.shape[1] != seq_length
                or position_ids.shape[0] not in (1, batch_size)
            ):
                raise ValueError(
                    "position_ids must be 1-D [seq] or 2-D [batch_or_1, seq] "
                    f"(got {position_ids.shape})."
                )

            if position_ids.shape[0] > 1:
                mismatch = (position_ids != position_ids[:1]).sum().long().item()
                if mismatch != 0:
                    raise ValueError(
                        "RoPE currently expects identical position_ids across batch."
                    )

            return position_ids[0].to(device).long()

        if cache_position is not None:
            if cache_position.ndim != 1 or cache_position.shape[0] != seq_length:
                raise ValueError(
                    "cache_position must be 1-D with length == seq_length "
                    f"(got {cache_position.shape}, seq_length={seq_length})."
                )
            return cache_position.to(device).long()

        past_key_values_length = 0
        if past_key_values is not None:
            current_past_key_value = past_key_values
            if isinstance(past_key_values, nn.EncoderDecoderCache):
                current_past_key_value = past_key_values.self_attention_cache

            if len(current_past_key_value.key_cache) > 0:
                past_key_values_length = current_past_key_value.get_seq_length()

        return lucid.arange(
            past_key_values_length,
            past_key_values_length + seq_length,
            device=device,
            dtype=lucid.Long,
        )

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        position_ids: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        del kwargs

        input_shape = hidden_states.shape[:-1]
        hidden_shape = (*input_shape, -1, self.attention_head_size)

        query_layer = self.query(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)
        key_layer = self.key(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)
        value_layer = self.value(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)

        rope_position_ids = self._resolve_rope_position_ids(
            hidden_states=hidden_states,
            position_ids=position_ids,
            cache_position=cache_position,
            past_key_values=past_key_values,
        )

        query_layer = self.rope_qk(query_layer, rope_position_ids)
        key_layer = self.rope_qk(key_layer, rope_position_ids)

        if self.rope_v is not None:
            value_layer = self.rope_v(value_layer, rope_position_ids)

        if past_key_values is not None:
            if self.layer_idx is None:
                raise ValueError("past_key_values requires layer_idx")

            current_past_key_value = past_key_values
            if isinstance(past_key_values, nn.EncoderDecoderCache):
                current_past_key_value = past_key_values.self_attention_cache

            key_layer, value_layer = current_past_key_value.update(
                key_layer,
                value_layer,
                layer_idx=self.layer_idx,
                cache_position=cache_position,
            )

        attn_output = F.scaled_dot_product_attention(
            query_layer,
            key_layer,
            value_layer,
            attn_mask=attention_mask,
            dropout_p=self.dropout.p if self.training else 0.0,
            is_causal=False,
            scale=self.scaling,
        )
        attn_output = attn_output.swapaxes(1, 2).reshape(
            *input_shape, self.all_head_size
        )
        return attn_output


class _RoFormerAttention(nn.Module):
    def __init__(
        self,
        config: RoFormerConfig,
        /,
        is_causal: bool = False,
        layer_idx: int | None = None,
        is_cross_attention: bool = False,
    ) -> None:
        super().__init__()
        self.is_cross_attention = is_cross_attention
        if is_cross_attention:
            self.self = bert._BERTCrossAttention(
                config, is_causal=is_causal, layer_idx=layer_idx
            )
        else:
            self.self = _RoFormerSelfAttention(config, layer_idx=layer_idx)
        self.output = bert._BERTSelfOutput(config)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        position_ids: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        del kwargs

        attention_mask = (
            attention_mask if not self.is_cross_attention else encoder_attention_mask
        )
        if (
            self.is_cross_attention
            and past_key_values is not None
            and not isinstance(past_key_values, nn.EncoderDecoderCache)
        ):
            raise TypeError(
                "Cross-attention requires past_key_values to be nn.EncoderDecoderCache."
            )

        attn_output = self.self(
            hidden_states,
            encoder_hidden_states=encoder_hidden_states,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            cache_position=cache_position,
            position_ids=position_ids,
        )
        attn_output = self.output(attn_output, hidden_states)
        return attn_output


class _RoFormerLayer(nn.Module):
    def __init__(self, config: RoFormerConfig, layer_idx: int | None = None) -> None:
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1

        self.attention = _RoFormerAttention(
            config, is_causal=config.is_decoder, layer_idx=layer_idx
        )
        self.is_decoder = config.is_decoder
        self.add_cross_attention = config.add_cross_attention

        if self.add_cross_attention:
            if not self.is_decoder:
                raise ValueError(
                    f"{self} should be used as a decoder if cross attention is added."
                )
            self.crossattention = _RoFormerAttention(
                config, is_causal=False, layer_idx=layer_idx, is_cross_attention=True
            )

        self.intermediate = bert._BERTIntermediate(config)
        self.output = bert._BERTOutput(config)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        position_ids: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        self_attention_output = self.attention(
            hidden_states,
            attention_mask,
            past_key_values=past_key_values,
            cache_position=cache_position,
            position_ids=position_ids,
            **kwargs,
        )
        attention_output = self_attention_output

        if self.is_decoder and encoder_hidden_states is not None:
            if not hasattr(self, "crossattention"):
                raise RuntimeError(
                    "If encoder_hidden_states is passed, "
                    "config.add_cross_attention must be set to True."
                )

            cross_attention_output = self.crossattention(
                self_attention_output,
                attention_mask=None,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                past_key_values=past_key_values,
                **kwargs,
            )
            attention_output = cross_attention_output

        layer_output = nn.utils.apply_chunking_to_forward(
            self.forward_chunk,
            self.chunk_size_feed_forward,
            self.seq_len_dim,
            attention_output,
        )
        return layer_output

    def forward_chunk(self, attention_output: Tensor) -> Tensor:
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class _RoFormerEncoder(nn.Module):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList(
            [
                _RoFormerLayer(config, layer_idx=i)
                for i in range(config.num_hidden_layers)
            ]
        )

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        use_cache: bool | None = None,
        cache_position: Tensor | None = None,
        position_ids: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        effective_use_cache = (
            use_cache if use_cache is not None else self.config.use_cache
        )
        if not self.config.is_decoder:
            effective_use_cache = False

        if (
            self.config.add_cross_attention
            and past_key_values is not None
            and not isinstance(past_key_values, nn.EncoderDecoderCache)
        ):
            raise TypeError(
                "When add_cross_attention=True, past_key_values must be "
                "nn.EncoderDecoderCache."
            )

        if not effective_use_cache:
            past_key_values = None
            cache_position = None
        elif past_key_values is None:
            raise ValueError(
                "use_cache=True requires a persistent past_key_values cache instance "
                "from the caller."
            )

        for layer_module in self.layer:
            hidden_states = layer_module(
                hidden_states,
                attention_mask,
                encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                past_key_values=past_key_values,
                cache_position=cache_position,
                position_ids=position_ids,
                **kwargs,
            )

        return hidden_states


class RoFormer(bert.BERT):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__(config)
        self.config = config

        self.embeddings = _RoFormerEmbeddings(config)
        self.encoder = _RoFormerEncoder(config)
        self.pooler = bert._BERTPooler(config) if config.add_pooling_layer else None
        self.output_embeddings: nn.Linear | None = None

        self.apply(self._init_weights)
        self.tie_weights()

    def forward(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: Tensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        use_cache: bool | None = None,
        cache_position: Tensor | None = None,
    ) -> tuple[Tensor, Tensor | None]:
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("Only one of input_ids or inputs_embeds can be provided.")
        if input_ids is None and inputs_embeds is None:
            raise ValueError("Either input_ids or inputs_embeds must be provided.")

        if input_ids is not None:
            input_shape = input_ids.shape
            device = input_ids.device
        else:
            input_shape = inputs_embeds.shape[:-1]
            device = inputs_embeds.device

        if len(input_shape) != 2:
            raise ValueError(
                f"RoFormer expects 2-D token inputs [batch, seq] (got {input_shape})."
            )

        effective_use_cache = (
            use_cache if use_cache is not None else self.config.use_cache
        )
        if not self.config.is_decoder:
            effective_use_cache = False

        if not effective_use_cache:
            past_key_values = None
            cache_position = None

        past_key_values_length = 0
        if effective_use_cache and past_key_values is not None:
            if isinstance(past_key_values, nn.EncoderDecoderCache):
                self_cache = past_key_values.self_attention_cache
                if len(self_cache.key_cache) > 0:
                    past_key_values_length = self_cache.get_seq_length()
            else:
                if len(past_key_values.key_cache) > 0:
                    past_key_values_length = past_key_values.get_seq_length()

        embedding_output = self.embeddings(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
            cache_position=cache_position,
        )

        source_length = input_shape[1] + past_key_values_length
        if attention_mask is None:
            attention_mask = lucid.ones((input_shape[0], source_length), device=device)

        elif attention_mask.ndim == 2 and attention_mask.shape[1] == input_shape[1]:
            if past_key_values_length > 0:
                past_attention = lucid.ones(
                    (input_shape[0], past_key_values_length), device=device
                )
                attention_mask = lucid.concatenate(
                    [past_attention, attention_mask], axis=1
                )

        extended_attention_mask = self._to_4d_attention_mask(
            attention_mask,
            target_length=input_shape[1],
            source_length=source_length,
            device=device,
        )

        if self.config.is_decoder:
            causal_attention_mask = self._build_decoder_causal_mask(
                batch_size=input_shape[0],
                target_length=input_shape[1],
                source_length=source_length,
                device=device,
                past_key_values_length=past_key_values_length,
                position_ids=position_ids,
                cache_position=cache_position,
            )
            if extended_attention_mask is None:
                extended_attention_mask = causal_attention_mask
            else:
                extended_attention_mask = (
                    extended_attention_mask + causal_attention_mask
                )

        if encoder_hidden_states is None:
            if encoder_attention_mask is not None:
                raise ValueError(
                    "encoder_attention_mask requires encoder_hidden_states."
                )
            extended_encoder_attention_mask = None
        else:
            extended_encoder_attention_mask = self._to_4d_attention_mask(
                attention_mask=encoder_attention_mask,
                target_length=input_shape[1],
                source_length=encoder_hidden_states.shape[1],
                device=device,
            )

        sequence_output = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=extended_encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=effective_use_cache,
            cache_position=cache_position,
            position_ids=position_ids,
        )
        pooled_output = (
            self.pooler(sequence_output) if self.pooler is not None else None
        )
        return sequence_output, pooled_output



from dataclasses import dataclass

import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

__all__ = ["Xception", "XceptionConfig", "xception"]


@dataclass
class XceptionConfig:
    num_classes: int = 1000
    in_channels: int = 3
    stem_channels: tuple[int, int] | list[int] = (32, 64)
    entry_channels: tuple[int, int, int] | list[int] = (128, 256, 728)
    middle_channels: int = 728
    middle_repeats: int = 8
    exit_channels: tuple[int, int, int] | list[int] = (1024, 1536, 2048)

    def __post_init__(self) -> None:
        self.stem_channels = tuple(self.stem_channels)
        self.entry_channels = tuple(self.entry_channels)
        self.exit_channels = tuple(self.exit_channels)

        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if len(self.stem_channels) != 2:
            raise ValueError("stem_channels must contain exactly 2 channel values")
        if any(
            not isinstance(channel, int) or channel <= 0
            for channel in self.stem_channels
        ):
            raise ValueError("stem_channels values must be positive integers")
        if len(self.entry_channels) != 3:
            raise ValueError("entry_channels must contain exactly 3 channel values")
        if any(
            not isinstance(channel, int) or channel <= 0
            for channel in self.entry_channels
        ):
            raise ValueError("entry_channels values must be positive integers")
        if self.middle_channels <= 0:
            raise ValueError("middle_channels must be greater than 0")
        if self.middle_repeats <= 0:
            raise ValueError("middle_repeats must be greater than 0")
        if len(self.exit_channels) != 3:
            raise ValueError("exit_channels must contain exactly 3 channel values")
        if any(
            not isinstance(channel, int) or channel <= 0
            for channel in self.exit_channels
        ):
            raise ValueError("exit_channels values must be positive integers")


class _Block(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        reps: int,
        stride: int = 1,
        start_with_relu: bool = True,
        grow_first: bool = True,
    ) -> None:
        super().__init__()
        if out_channels != in_channels or stride != 1:
            self.skip = nn.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=stride, bias=False
            )
            self.skipbn = nn.BatchNorm2d(out_channels)
        else:
            self.skip = None

        rep = []
        channels = in_channels
        if grow_first:
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    in_channels, out_channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(out_channels))
            channels = out_channels

        for _ in range(reps - 1):
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    channels, channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(channels))

        if not grow_first:
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    in_channels, out_channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(out_channels))

        if not start_with_relu:
            rep = rep[1:]
        else:
            rep[0] = nn.ReLU()

        if stride != 1:
            rep.append(nn.MaxPool2d(kernel_size=3, stride=stride, padding=1))

        self.rep = nn.Sequential(*rep)

    def forward(self, x: Tensor) -> Tensor:
        out = self.rep(x)

        if self.skip is not None:
            skip = self.skip(x)
            skip = self.skipbn(skip)
        else:
            skip = x

        out += skip
        return out


class Xception(nn.Module):
    def __init__(self, config: XceptionConfig) -> None:
        super().__init__()
        self.config = config
        self.relu = nn.ReLU()

        self.conv1 = nn.ConvBNReLU2d(
            config.in_channels,
            config.stem_channels[0],
            kernel_size=3,
            stride=2,
            padding=0,
            conv_bias=False,
        )
        self.conv2 = nn.ConvBNReLU2d(
            config.stem_channels[0],
            config.stem_channels[1],
            kernel_size=3,
            conv_bias=False,
        )

        self.block1 = _Block(
            config.stem_channels[1],
            config.entry_channels[0],
            reps=2,
            stride=2,
            start_with_relu=False,
        )
        self.block2 = _Block(
            config.entry_channels[0], config.entry_channels[1], reps=2, stride=2
        )
        self.block3 = _Block(
            config.entry_channels[1], config.entry_channels[2], reps=2, stride=2
        )

        self.mid_blocks = nn.Sequential(
            *[
                _Block(config.middle_channels, config.middle_channels, reps=3)
                for _ in range(config.middle_repeats)
            ]
        )
        self.end_block = _Block(
            config.middle_channels,
            config.exit_channels[0],
            reps=2,
            stride=2,
            grow_first=False,
        )

        self.conv3 = nn.DepthSeparableConv2d(
            config.exit_channels[0],
            config.exit_channels[1],
            kernel_size=3,
            padding=1,
        )
        self.bn3 = nn.BatchNorm2d(config.exit_channels[1])

        self.conv4 = nn.DepthSeparableConv2d(
            config.exit_channels[1],
            config.exit_channels[2],
            kernel_size=3,
            padding=1,
        )
        self.bn4 = nn.BatchNorm2d(config.exit_channels[2])

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(config.exit_channels[2], config.num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)

        x = self.block3(self.block2(self.block1(x)))
        x = self.mid_blocks(x)
        x = self.end_block(x)

        x = self.relu(self.bn3(self.conv3(x)))
        x = self.relu(self.bn4(self.conv4(x)))

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


def _build_xception_config(
    *,
    num_classes: int,
    kwargs: dict[str, object] | None = None,
) -> XceptionConfig:
    kwargs = {} if kwargs is None else dict(kwargs)
    locked_fields = {
        "stem_channels",
        "entry_channels",
        "middle_channels",
        "middle_repeats",
        "exit_channels",
    }
    if locked_fields & kwargs.keys():
        raise TypeError(
            "factory variants do not allow overriding preset stem_channels, entry_channels, middle_channels, middle_repeats, or exit_channels"
        )

    return XceptionConfig(num_classes=num_classes, **kwargs)


@register_model
def xception(num_classes: int = 1000, **kwargs) -> Xception:
    config = _build_xception_config(num_classes=num_classes, kwargs=kwargs)
    return Xception(config)



from dataclasses import dataclass

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = ["GPTConfig", "GPT"]


@dataclass
class GPTConfig:
    vocab_size: int = 40478
    max_position_embeddings: int = 512

    hidden_size: int = 768
    num_attention_heads: int = 12
    num_hidden_layers: int = 12

    intermediate_size: int = 3072
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.1

    attention_prob_dropout_prob: float = 0.1
    layer_norm_eps: float = 1e-5
    initializer_range: float = 0.02

    use_cache: bool = True
    bos_token_id: int = 40476
    eos_token_id: int = 40477
    pad_token_id: int | None = None

    def __post_init__(self) -> None:
        assert (
            self.hidden_size % self.num_attention_heads == 0
        ), "hidden_size must be divisible by num_attention_heads"
        assert self.hidden_size > 0 and self.num_hidden_layers > 0

    @classmethod
    def base(cls, **kwargs) -> GPTConfig:
        defaults = dict(
            hidden_size=768,
            num_attention_heads=12,
            num_hidden_layers=12,
            intermediate_size=3072,
        )
        defaults.update(kwargs)
        return cls(**defaults)


class _GPTEmbedding(nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.tokens_embed = nn.Embedding(config.vocab_size, config.hidden_size)
        self.positions_embed = nn.Embedding(
            config.max_position_embeddings, config.hidden_size
        )

        self.drop = nn.Dropout(config.hidden_dropout_prob)
        self.register_buffer(
            "position_ids",
            lucid.arange(config.max_position_embeddings).unsqueeze(axis=0),
        )
        self.position_ids: nn.Buffer

    def forward(
        self,
        input_ids: Tensor,
        position_ids: Tensor | None = None,
        past_length: int = 0,
    ) -> Tensor:
        seq_len = input_ids.shape[1]
        if position_ids is None:
            position_ids = self.position_ids[:, past_length : past_length + seq_len]
        return self.drop(
            self.tokens_embed(input_ids) + self.positions_embed(position_ids)
        )


class _GPTAttention(nn.Module):
    def __init__(self, config: GPTConfig, layer_idx: int = 0) -> None:
        super().__init__()
        self.num_heads = config.num_attention_heads
        self.head_dim = config.hidden_size // config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.layer_idx = layer_idx
        self.scale = self.head_dim**-0.5

        self.c_attn = nn.Linear(config.hidden_size, 3 * config.hidden_size)
        self.c_proj = nn.Linear(config.hidden_size, config.hidden_size)
        self.attn_drop = nn.Dropout(config.attention_prob_dropout_prob)
        self.resid_drop = nn.Dropout(config.hidden_dropout_prob)

        max_len = config.max_position_embeddings
        self.register_buffer(
            "causal_mask",
            lucid.tril(lucid.ones(max_len, max_len)).reshape(1, 1, max_len, max_len),
        )
        self.causal_mask: nn.Buffer

    def forward(
        self,
        hidden_state: Tensor,
        attention_mask: Tensor | None = None,
        past_key_value: nn.KVCache | None = None,
        cache_position: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor, nn.KVCache | None]:
        B, T, C = hidden_state.shape
        q, k, v = self.c_attn(hidden_state).split(self.hidden_size, axis=-1)

        q = q.reshape(B, T, self.num_heads, self.head_dim).swapaxes(1, 2)
        k = k.reshape(B, T, self.num_heads, self.head_dim).swapaxes(1, 2)
        v = v.reshape(B, T, self.num_heads, self.head_dim).swapaxes(1, 2)

        if past_key_value is not None:
            past_len = past_key_value.get_seq_length(self.layer_idx)
            k, v = past_key_value.update(k, v, self.layer_idx, cache_position)
        else:
            past_len = 0

        kv_len = k.shape[-2]
        mask = self.causal_mask[:, :, past_len : past_len + T, :kv_len]
        if attention_mask is not None:
            mask = mask + attention_mask

        attn_out = F.scaled_dot_product_attention(
            q,
            k,
            v,
            attn_mask=mask,
            dropout_p=self.attn_drop.p if self.training else 0.0,
            scale=self.scale,
        )
        attn_out = attn_out.swapaxes(1, 2).reshape(B, T, C)
        attn_out = self.resid_drop(self.c_proj(attn_out))

        return attn_out, past_key_value if use_cache else None


class _GPTMLP(nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.c_fc = nn.Linear(config.hidden_size, config.intermediate_size)
        self.c_proj = nn.Linear(config.intermediate_size, config.hidden_size)

        self.drop = nn.Dropout(config.hidden_dropout_prob)
        self.act = nn.utils.get_activation_module_from_name(config.hidden_act)
        if self.act is None:
            raise ValueError(f"Invalid activation name: '{config.hidden_act}'")

    def forward(self, x: Tensor) -> Tensor:
        return self.drop(self.c_proj(self.act(self.c_fc(x))))


class _GPTBlock(nn.Module):
    def __init__(self, config: GPTConfig, layer_idx: int = 0) -> None:
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.attn = _GPTAttention(config, layer_idx=layer_idx)
        self.ln_2 = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.mlp = _GPTMLP(config)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor | None = None,
        past_key_value: nn.KVCache | None = None,
        cache_position: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor, nn.KVCache | None]:
        attn_out, present = self.attn(
            self.ln_1(hidden_states),
            attention_mask=attention_mask,
            past_key_value=past_key_value,
            cache_position=cache_position,
            use_cache=use_cache,
        )
        hidden_states = hidden_states + attn_out
        hidden_states = hidden_states + self.mlp(self.ln_2(hidden_states))

        return hidden_states, present


class _GPTDecoder(nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.h = nn.ModuleList(
            [_GPTBlock(config, layer_idx=i) for i in range(config.num_hidden_layers)]
        )

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor | None = None,
        past_key_values: list[nn.KVCache] | None = None,
        cache_position: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor, list[nn.KVCache] | None]:
        presents: list[nn.KVCache] | None = [] if use_cache else None
        for i, block in enumerate(self.h):
            pkv = past_key_values[i] if past_key_values is not None else None
            hidden_states, present = block(
                hidden_states,
                attention_mask=attention_mask,
                past_key_value=pkv,
                cache_position=cache_position,
                use_cache=use_cache,
            )
            if use_cache:
                presents.append(present)

        return hidden_states, presents


class GPT(PreTrainedModelMixin, nn.Module):
    def __init__(self, config: GPTConfig) -> None:
        super().__init__()
        self.config = config
        self.embeddings = _GPTEmbedding(config)
        self.decoder = _GPTDecoder(config)

        self.apply(self._init_weights)

    def _init_weights(self, module: nn.Module) -> None:
        if isinstance(module, nn.Linear):
            nn.init.normal(module.weight, mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                nn.init.constant(module.bias, 0.0)

        elif isinstance(module, nn.Embedding):
            nn.init.normal(module.weight, mean=0.0, std=self.config.initializer_range)

        elif isinstance(module, nn.LayerNorm):
            nn.init.constant(module.bias, 0.0)
            nn.init.constant(module.weight, 1.0)

    def get_input_embeddings(self) -> nn.Embedding:
        return self.embeddings.tokens_embed

    def set_input_embeddings(self, value: nn.Embedding) -> None:
        self.embeddings.tokens_embed = value

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Tensor | None = None,
        position_ids: Tensor | None = None,
        past_key_values: list[nn.KVCache] | None = None,
        cache_position: Tensor | None = None,
        use_cache: bool = False,
    ) -> tuple[Tensor, list[nn.KVCache] | None]:
        past_length = (
            past_key_values[0].get_seq_length() if past_key_values is not None else 0
        )
        hidden_states = self.embeddings(input_ids, position_ids, past_length)
        hidden_states, presents = self.decoder(
            hidden_states,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            cache_position=cache_position,
            use_cache=use_cache,
        )
        return hidden_states, presents



from dataclasses import dataclass, field
from typing import Any, Literal

from lucid import register_model
from .resnet import ResNet, ResNetConfig
from lucid.models.base import PreTrainedModelMixin

__all__ = [
    "ResNeXt",
    "ResNeXtConfig",
    "resnext_50_32x4d",
    "resnext_101_32x4d",
    "resnext_101_32x8d",
    "resnext_101_32x16d",
    "resnext_101_32x32d",
    "resnext_101_64x4d",
]


@dataclass
class ResNeXtConfig:
    layers: tuple[int, int, int, int] | list[int]
    cardinality: int
    base_width: int
    num_classes: int = 1000
    in_channels: int = 3
    stem_width: int = 64
    stem_type: Literal["deep"] | None = None
    avg_down: bool = False
    channels: tuple[int, int, int, int] | list[int] = (64, 128, 256, 512)
    block_args: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.layers = tuple(self.layers)
        self.channels = tuple(self.channels)
        if not isinstance(self.block_args, dict):
            raise TypeError("block_args must be a dictionary")
        self.block_args = dict(self.block_args)

        if len(self.layers) != 4:
            raise ValueError("layers must contain exactly 4 stage depths")
        if any(not isinstance(depth, int) or depth <= 0 for depth in self.layers):
            raise ValueError("layers values must be positive integers")
        if self.cardinality <= 0:
            raise ValueError("cardinality must be greater than 0")
        if self.base_width <= 0:
            raise ValueError("base_width must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if self.stem_width <= 0:
            raise ValueError("stem_width must be greater than 0")
        if self.stem_type not in (None, "deep"):
            raise ValueError("stem_type must be None or 'deep'")
        if len(self.channels) != 4:
            raise ValueError("channels must contain exactly 4 stage widths")
        if any(
            not isinstance(channel, int) or channel <= 0 for channel in self.channels
        ):
            raise ValueError("channels values must be positive integers")


class ResNeXt(ResNet, PreTrainedModelMixin):
    def __init__(self, config: ResNeXtConfig) -> None:
        block_args = {
            "cardinality": config.cardinality,
            "base_width": config.base_width,
            **config.block_args,
        }
        super().__init__(
            ResNetConfig(
                block="bottleneck",
                layers=config.layers,
                num_classes=config.num_classes,
                in_channels=config.in_channels,
                stem_width=config.stem_width,
                stem_type=config.stem_type,
                avg_down=config.avg_down,
                channels=config.channels,
                block_args=block_args,
            )
        )
        self.config = config
        self.cardinality = config.cardinality
        self.base_width = config.base_width


def _build_resnext_config(
    *,
    layers: tuple[int, int, int, int] | list[int],
    cardinality: int,
    base_width: int,
    num_classes: int,
    kwargs: dict[str, Any] | None = None,
) -> ResNeXtConfig:
    kwargs = {} if kwargs is None else dict(kwargs)
    locked_fields = {"layers", "cardinality", "base_width"}
    if locked_fields & kwargs.keys():
        raise TypeError(
            "factory variants do not allow overriding preset layers, cardinality, or base_width"
        )

    return ResNeXtConfig(
        layers=layers,
        cardinality=cardinality,
        base_width=base_width,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def resnext_50_32x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 6, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=32,
        base_width=4,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)


@register_model
def resnext_101_32x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=32,
        base_width=4,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)


@register_model
def resnext_101_32x8d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=32,
        base_width=8,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)


@register_model
def resnext_101_32x16d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=32,
        base_width=16,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)


@register_model
def resnext_101_32x32d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=32,
        base_width=32,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)


@register_model
def resnext_101_64x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    config = _build_resnext_config(
        layers=layers,
        cardinality=64,
        base_width=4,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeXt(config)



from __future__ import annotations
import collections.abc
import typing

__all__: list[str] = list()

class _C_BPETokenizer:
    def __init__(
        self,
        vocab: collections.abc.Mapping[str, typing.SupportsInt] | None = None,
        merges: collections.abc.Sequence[tuple[str, str]] | None = None,
        vocab_file: os.PathLike | str | bytes | None = None,
        merges_file: os.PathLike | str | bytes | None = None,
        unk_token: str = "[UNK]",
        pad_token: str = "[PAD]",
        bos_token: str | None = None,
        eos_token: str | None = None,
        lowercase: bool = True,
        clean_text: bool = True,
        end_of_word_suffix: str = "</w>",
    ) -> None: ...
    def convert_id_to_token(self, id: typing.SupportsInt) -> str: ...
    def convert_ids_to_tokens(
        self, ids: collections.abc.Sequence[typing.SupportsInt]
    ) -> list[str]: ...
    def convert_token_to_id(self, token: str) -> int: ...
    def convert_tokens_to_ids(
        self, tokens: collections.abc.Sequence[str]
    ) -> list[int]: ...
    def convert_tokens_to_string(
        self, tokens: collections.abc.Sequence[str]
    ) -> str: ...
    def encode_ids(self, text: str, add_special_tokens: bool = True) -> list[int]: ...
    def fit(
        self,
        texts: collections.abc.Sequence[str],
        vocab_size: typing.SupportsInt,
        min_frequency: typing.SupportsInt = 2,
    ) -> _C_BPETokenizer: ...
    def get_merges(self) -> list[tuple[str, str]]: ...
    def get_vocab(self) -> dict[str, int]: ...
    def tokenize(self, text: str) -> list[str]: ...
    def vocab_size(self) -> int: ...

class _C_ByteBPETokenizer(_C_BPETokenizer):
    def __init__(
        self,
        vocab: collections.abc.Mapping[str, typing.SupportsInt] | None = None,
        merges: collections.abc.Sequence[tuple[str, str]] | None = None,
        vocab_file: os.PathLike | str | bytes | None = None,
        merges_file: os.PathLike | str | bytes | None = None,
        unk_token: str = "[UNK]",
        pad_token: str = "[PAD]",
        bos_token: str | None = None,
        eos_token: str | None = None,
        lowercase: bool = False,
        clean_text: bool = True,
        add_prefix_space: bool = False,
        end_of_word_suffix: str = "",
    ) -> None: ...
    def add_prefix_space(self) -> bool: ...
    def convert_id_to_token(self, id: typing.SupportsInt) -> str: ...
    def convert_ids_to_tokens(
        self, ids: collections.abc.Sequence[typing.SupportsInt]
    ) -> list[str]: ...
    def convert_token_to_id(self, token: str) -> int: ...
    def convert_tokens_to_ids(
        self, tokens: collections.abc.Sequence[str]
    ) -> list[int]: ...
    def convert_tokens_to_string(
        self, tokens: collections.abc.Sequence[str]
    ) -> str: ...
    def encode_ids(self, text: str, add_special_tokens: bool = True) -> list[int]: ...
    def fit(
        self,
        texts: collections.abc.Sequence[str],
        vocab_size: typing.SupportsInt,
        min_frequency: typing.SupportsInt = 2,
    ) -> _C_ByteBPETokenizer: ...
    def get_merges(self: _C_BPETokenizer) -> list[tuple[str, str]]: ...
    def get_vocab(self: _C_BPETokenizer) -> dict[str, int]: ...
    def tokenize(self, text: str) -> list[str]: ...
    def vocab_size(self) -> int: ...

class _C_WordPieceTokenizer:
    def __init__(
        self,
        vocab: collections.abc.Mapping[str, typing.SupportsInt] | None = None,
        vocab_file: os.PathLike | str | bytes | None = None,
        unk_token: str = "[UNK]",
        pad_token: str = "[PAD]",
        bos_token: str | None = None,
        eos_token: str | None = None,
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: typing.SupportsInt = 100,
        lowercase: bool = True,
        clean_text: bool = True,
    ) -> None: ...
    def convert_id_to_token(self, id: typing.SupportsInt) -> str: ...
    def convert_ids_to_tokens(
        self, ids: collections.abc.Sequence[typing.SupportsInt]
    ) -> list[str]: ...
    def convert_token_to_id(self, token: str) -> int: ...
    def convert_tokens_to_ids(
        self, tokens: collections.abc.Sequence[str]
    ) -> list[int]: ...
    def convert_tokens_to_string(
        self, tokens: collections.abc.Sequence[str]
    ) -> str: ...
    def fit(
        self,
        texts: collections.abc.Sequence[str],
        vocab_size: typing.SupportsInt,
        min_frequency: typing.SupportsInt = 2,
    ) -> _C_WordPieceTokenizer: ...
    def tokenize(self, text: str) -> list[str]: ...
    def vocab_size(self) -> int: ...



from abc import ABC, abstractmethod
from typing import Callable, Iterator, Self, Any, override
import random
import math

import lucid
from lucid._tensor import Tensor
from lucid.types import _ArrayLike, _IndexLike, _DeviceType


class Dataset(ABC):
    @abstractmethod
    def __getitem__(self, idx: _IndexLike) -> None:
        raise NotImplementedError("Subclasses must implement __getitem__.")

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError("Subclasses must implement __len__.")

    def __add__(self, other: Self) -> Self:
        return ConcatDataset([self, other])

    def __iter__(self) -> Iterator[Any]:
        for i in range(len(self)):
            yield self[i]

    def __repr__(self) -> str:
        return f"Dataset(n={len(self)})"


class Subset(Dataset):
    def __init__(self, dataset: Dataset, indices: list[int]) -> None:
        super().__init__()
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx: _IndexLike) -> Any:
        return self.dataset[self.indices[idx]]

    def __len__(self) -> int:
        return len(self.indices)

    @override
    def __iter__(self) -> Iterator[Any]:
        for i in self.indices:
            yield self.dataset[i]

    def __getattr__(self, name: str) -> Any:
        return getattr(self.dataset, name)

    def __repr__(self) -> str:
        return f"Subset(n={len(self)})"


class TensorDataset(Dataset):
    def __init__(self, *tensors_or_arrays: Tensor | _ArrayLike) -> None:
        super().__init__()
        if len(tensors_or_arrays) == 0:
            raise ValueError(
                "TensorDataset requires at least one tensor/array-like object."
            )
        try:
            self._tensors: tuple[Tensor, ...] = tuple(
                lucid._check_is_tensor(t) for t in tensors_or_arrays
            )
        except Exception as e:
            raise RuntimeError(
                "Failed to convert array-like object(s) to tensor."
            ) from e

        n0 = len(self._tensors[0])
        for i, t in enumerate(self._tensors):
            if t.ndim == 0 or len(t) == 0:
                raise RuntimeError(
                    "All tensors must be at least 1D. "
                    f"Tensor at index {i} has no length."
                )
            if len(t) != n0:
                raise ValueError(
                    "All tensors must have the same length along dim 0: "
                    f"got {n0} and {len(t)} at index {i}."
                )

    def __len__(self) -> int:
        return len(self._tensors[0])

    @override
    def __getitem__(self, idx: _IndexLike | Tensor) -> tuple[Tensor, ...]:
        return tuple(t[idx] for t in self._tensors)

    def to(self, device: _DeviceType) -> Self:
        self._tensors = tuple(t.to(device) for t in self._tensors)
        return self

    @property
    def tensors(self) -> tuple[Tensor, ...]:
        return self._tensors

    @override
    def __iter__(self) -> Iterator[tuple[Tensor, ...]]:
        return super().__iter__()

    @override
    def __repr__(self) -> str:
        shapes = ", ".join(str(t.shape) for t in self._tensors)
        devices = {t.device for t in self._tensors}
        return f"TensorDataset(n={len(self)}, shapes=({shapes}), devices={devices})"


class ConcatDataset(Dataset):
    def __init__(self, datasets: list[Dataset]) -> None:
        super().__init__()
        self.datasets = datasets
        self.cumulative_sizes = self._compute_cumulative_sizes()

    def _compute_cumulative_sizes(self) -> list[int]:
        cum_sizes = []
        total = 0
        for dataset in self.datasets:
            total += len(dataset)
            cum_sizes.append(total)

        return cum_sizes

    def __len__(self) -> int:
        return self.cumulative_sizes[-1] if self.cumulative_sizes else 0

    def __getitem__(self, idx: _IndexLike) -> Any:
        if idx < 0:
            if -idx > len(self):
                raise IndexError("Index out of range.")
            idx = len(self) + idx

        for i, size in enumerate(self.cumulative_sizes):
            if idx < size:
                dataset_idx = i
                if dataset_idx > 0:
                    idx -= self.cumulative_sizes[dataset_idx - 1]

                return self.datasets[dataset_idx][idx]

        raise IndexError("Index out of range.")


class DataLoader:
    def __init__(
        self,
        dataset: Dataset,
        batch_size: int = 1,
        shuffle: bool = False,
        collate_fn: Callable | None = None,
    ) -> None:
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.collate_fn = collate_fn or self.default_collate
        self.indices = list(range(len(dataset)))

        if shuffle:
            self._shuffle_indices()

    def _shuffle_indices(self) -> None:
        random.shuffle(self.indices)

    def __iter__(self) -> Self:
        self.current_index = 0
        if self.shuffle:
            self._shuffle_indices()
        return self

    def __next__(self) -> Any:
        if self.current_index >= len(self.indices):
            raise StopIteration

        start = self.current_index
        end = min(start + self.batch_size, len(self.indices))
        batch_indices = self.indices[start:end]
        self.current_index = end

        batch = [self.dataset[i] for i in batch_indices]
        return self.collate_fn(batch)

    def __len__(self) -> int:
        return int(math.ceil(len(self.dataset) / self.batch_size))

    @staticmethod
    def default_collate(batch: list[Any]) -> Any:
        if isinstance(batch[0], (tuple, list)):
            transposed = list(zip(*batch))
            return tuple(lucid.stack(tuple(x), axis=0) for x in transposed)

        elif isinstance(batch[0], Tensor):
            return lucid.stack(tuple(batch), axis=0)
        else:
            return batch



from dataclasses import dataclass, field
from typing import Any, Callable

from lucid.types import _DeviceType


@dataclass
class ShapeSpec:
    shape: tuple[int, ...]
    dtype: Any
    device: _DeviceType


@dataclass
class IRValue:
    value_id: int
    shape_spec: ShapeSpec
    live_tensor: Any = field(default=None, repr=False)


@dataclass
class IRNode:
    node_id: int
    op_class: type
    op_init_args: tuple
    op_init_kwargs: dict

    input_ids: list[int]
    output_ids: list[int]

    non_tensor_args: tuple
    non_tensor_kwargs: dict

    device: _DeviceType
    has_gradient: bool
    n_in: int | None
    n_ret: int | None

    fused_forward_kernel: Callable | None = None


@dataclass
class IRGraph:
    input_ids: list[int]
    output_ids: list[int]
    values: dict[int, IRValue]

    nodes: list[IRNode]
    param_ids: set[int]

    @property
    def constant_ids(self) -> set[int]:
        produced: set[int] = set()
        for node in self.nodes:
            produced.update(node.output_ids)
        return set(self.values.keys()) - set(self.input_ids) - self.param_ids - produced



from .core import ModuleImportBase


class TestPublicModulesImport(ModuleImportBase):
    modules = (
        "lucid",
        "lucid.types",
        "lucid.error",
        "lucid._tensor",
        "lucid._utils",
        "lucid._backend",
        "lucid._backend.metal",
        "lucid._func",
        "lucid.autograd",
        "lucid.random",
        "lucid.linalg",
        "lucid.nn",
        "lucid.optim",
        "lucid.models",
        "lucid.transforms",
        "lucid.visual",
        "lucid.weights",
        "lucid.data",
        "lucid.datasets",
        "lucid.einops",
    )



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, OrderedDict

import hashlib
import json
import os
import re
import shutil
import ssl
import tempfile
import urllib.error
import urllib.request
from urllib.parse import urlparse

import lucid
import lucid.nn as nn

from lucid._tensor import Tensor

_WEIGHTS_DIR: Path = Path(__file__).resolve().parent
WEIGHTS_REGISTRY_PATH: Path = _WEIGHTS_DIR / "registry.json"

CACHE_HOME = (
    Path(os.environ.get("LUCID_HOME", Path.home() / ".cache" / "lucid")) / "weights"
)
BASE_URL = os.environ.get("LUCID_WEIGHTS_BASE", "")
MIRRORS = [
    s for s in os.environ.get("LUCID_WEIGHTS_MIRRORS", "").split(",") if s.strip()
]
OFFLINE = os.environ.get("LUCID_WEIGHTS_OFFLINE", "0") == "1"
HF_ENDPOINT = os.environ.get("HF_ENDPOINT", "https://huggingface.co")
HF_TOKEN = (
    os.environ.get("LUCID_HF_TOKEN")
    or os.environ.get("HF_TOKEN")
    or os.environ.get("HUGGINGFACEHUB_API_TOKEN")
)
SSL_NO_VERIFY = os.environ.get("LUCID_SSL_NO_VERIFY", "0") == "1"
SSL_CA_BUNDLE = (
    os.environ.get("LUCID_SSL_CA_BUNDLE")
    or os.environ.get("SSL_CERT_FILE")
    or os.environ.get("REQUESTS_CA_BUNDLE")
)


def _build_ssl_context() -> ssl.SSLContext:
    if SSL_NO_VERIFY:
        c = ssl._create_unverified_context()
        c.check_hostname = False
        c.verify_mode = ssl.CERT_NONE
        return c

    if SSL_CA_BUNDLE:
        return ssl.create_default_context(cafile=SSL_CA_BUNDLE)

    try:
        import certifi

        return ssl.create_default_context(cafile=certifi.where())
    except Exception:
        return ssl.create_default_context()


_SSL_CONTEXT = _build_ssl_context()
SHOW_PROGRESS = os.environ.get("LUCID_WEIGHTS_PROGRESS", "1") != "0"

try:
    from tqdm.auto import tqdm
except Exception:
    tqdm = None


@dataclass(frozen=True)
class WeightEntry:
    url: str
    sha256: str
    tag: str
    dataset: str | None = None
    meta: dict[str, Any] | None = None

    @property
    def config(self) -> dict[str, Any] | None:
        if not isinstance(self.meta, dict):
            return None
        cfg = self.meta.get("config")
        if isinstance(cfg, dict):
            return cfg
        return None


def _load_json(path: Path) -> dict[str, Any]:
    if not path.exists():
        return {}
    with open(path, "r") as f:
        return json.load(f)


_REGISTRY = _load_json(WEIGHTS_REGISTRY_PATH)
_MODELS_META = _load_json(lucid.MODELS_REGISTRY_PATH)
_PYI_PATH = _WEIGHTS_DIR / "__init__.pyi"

_DYNAMIC: dict[str, Enum] = {}


def _load_stub_classnames() -> set[str]:
    if not _PYI_PATH.exists():
        return set()
    try:
        text = _PYI_PATH.read_text(encoding="utf-8")
    except Exception:
        return set()
    return set(
        re.findall(r"^class\s+([A-Za-z0-9_]+)\(Enum\):", text, flags=re.MULTILINE)
    )


_STUB_CLASSNAMES = _load_stub_classnames()


def _norm_key(name: str) -> str:
    return "".join(ch for ch in name.lower() if ch.isalnum())


def _build_stub_modelkey_map(classnames: set[str]) -> dict[str, str]:
    mapping: dict[str, str] = {}
    for cls in classnames:
        if not cls.endswith("_Weights"):
            continue
        model_part = cls[: -len("_Weights")]
        mapping.setdefault(_norm_key(model_part), cls)
    return mapping


_STUB_MODELKEY_MAP = _build_stub_modelkey_map(_STUB_CLASSNAMES)


def _canon(x: str) -> str:
    return x.lower()


def _find_model_meta(node: dict[str, Any], model_key: str) -> dict[str, Any] | None:
    if model_key in node and isinstance(node[model_key], dict):
        return node[model_key]

    for value in node.values():
        if not isinstance(value, dict):
            continue
        found = _find_model_meta(value, model_key)
        if found is not None:
            return found
    return None


def _family_of(model_key: str) -> str | None:
    meta = _find_model_meta(_MODELS_META, model_key)
    if isinstance(meta, dict):
        return meta.get("family")
    return None


def _classname(
    model_key: str, family: str | None = None, enum_name: str | None = None
) -> str:
    if enum_name:
        return enum_name

    stub_match = _STUB_MODELKEY_MAP.get(_norm_key(model_key))
    if stub_match is not None:
        return stub_match

    candidates: list[str] = []

    if family is not None:
        candidates.append(family + model_key[len(family) :].upper() + "_Weights")
    else:
        fam = _family_of(model_key)
        if fam:
            suf = model_key
            fam_key = fam.lower().replace(" ", "")
            suf = (
                suf[len(fam_key) :] if suf.replace("_", "").startswith(fam_key) else suf
            )
            suf = suf.lstrip()
            candidates.append(
                f"{fam}{suf.upper()}_Weights" if suf else f"{fam}_Weights"
            )

        parts = model_key.split("_")
        candidates.append("_".join(p.capitalize() for p in parts) + "_Weights")

    for name in candidates:
        if name in _STUB_CLASSNAMES:
            return name

    return candidates[0]


def _make_enum(model_key: str, entries: dict[str, Any]) -> type[Enum]:
    default_meta = entries.get("DEFAULT", {}).get("meta", {})
    members = {}
    for tag, info in entries.items():
        members[tag] = WeightEntry(
            url=info["url"],
            sha256=info["sha256"],
            tag=tag,
            dataset=info.get("dataset"),
            meta=info.get("meta", {}),
        )

    cl = Enum(
        _classname(
            model_key,
            family=default_meta.get("family", None),
            enum_name=default_meta.get("enum_name", None),
        ),
        members,
    )
    if not hasattr(cl, "config"):
        setattr(cl, "config", property(lambda self: self.value.config))
    _DYNAMIC[_canon(model_key)] = cl
    return cl


_created_names: list[str] = []
for k, entries in _REGISTRY.items():
    cl = _make_enum(k, entries)
    globals()[cl.__name__] = cl
    _created_names.append(cl.__name__)


def get_enum(model_key: str) -> type[Enum]:
    key = _canon(model_key)
    if key not in _DYNAMIC:
        raise KeyError(f"No weights registered for '{model_key}'.")
    return _DYNAMIC[key]


def resolve(weights: Enum | str, *, model_key: str | None = None) -> WeightEntry:
    if isinstance(weights, Enum):
        return weights.value
    if model_key is None:
        raise ValueError("model_key is required when resolving from string.")

    enum = get_enum(model_key)
    try:
        return getattr(enum, weights).value
    except AttributeError:
        return getattr(enum, "DEFAULT").value


def _hash_file(path: Path) -> str:
    h = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            h.update(chunk)
    return h.hexdigest()


def _hf_build_url(repo_id: str, revision: str, filename: str) -> str:
    rid = repo_id.strip("/")
    rev = revision or "main"
    return f"{HF_ENDPOINT.rstrip('/')}/{rid}/resolve/{rev}/{filename}?download=true"


def _parse_hf_uri(uri: str) -> tuple[str, str, str]:
    if not uri.startswith("hf://"):
        raise ValueError("Not an hf:// URI")

    s = uri[5:].lstrip("/")
    parts = s.split("/")
    if len(parts) < 3:
        raise ValueError("HF URI must be 'hf://owner/repo[@rev]/path'")

    owner = parts[0]
    repo_and_rev = parts[1]
    file_path = "/".join(parts[2:])

    if "@" in repo_and_rev:
        repo, rev = repo_and_rev.split("@", 1)
    else:
        repo, rev = repo_and_rev, "main"

    repo_id = f"{owner}/{repo}"
    return repo_id, rev, file_path


def _hf_candidate_urls(entry: WeightEntry) -> list[str]:
    urls: list[str] = []
    try:
        if entry.url.startswith("hf://"):
            repo_id, rev, file_path = _parse_hf_uri(entry.url)
            urls.append(_hf_build_url(repo_id, rev, file_path))
    except Exception:
        pass

    m = entry.meta or {}
    repo = m.get("hf_repo")
    file = m.get("hf_filename") or m.get("hf_path")
    rev = m.get("hf_revision", "main")

    if repo and file:
        urls.append(_hf_build_url(repo, rev, file))
    return urls


def _candidate_urls(entry: WeightEntry) -> list[str]:
    urls = []
    urls.extend(_hf_candidate_urls(entry))

    if entry.url and not entry.url.startswith("hf://"):
        urls.append(entry.url)
        if BASE_URL and entry.url.startswith("http"):
            tail = "/".join(entry.url.split("/")[3:])
            urls.insert(0, BASE_URL.rstrip("/") + "/" + tail)

        for m in MIRRORS:
            m = m.strip()
            if not m:
                continue
            if entry.url.startswith("http"):
                tail = "/".join(entry.url.split("/")[3:])
                urls.append(m.rstrip("/") + "/" + tail)
            else:
                urls.append(m.rstrip("/") + "/" + entry.url.lstrip("/"))

    return urls


def _cache_path(model_key: str, entry: WeightEntry) -> Path:
    fname = f"{_canon(model_key)}--{entry.tag}--{entry.sha256[:8]}.safetensors"
    return CACHE_HOME / fname


def _download_with_progress(resp, out, *, desc: str) -> None:
    total = 0
    try:
        total = int(resp.headers.get("Content-Length", "0"))
    except Exception:
        total = 0

    chunk_size = 1024 * 1024
    if tqdm is None or not SHOW_PROGRESS:
        shutil.copyfileobj(resp, out, length=chunk_size)
        return

    with tqdm(
        total=total if total > 0 else None,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        desc=desc,
        leave=False,
    ) as pbar:
        while True:
            chunk = resp.read(chunk_size)
            if not chunk:
                break
            out.write(chunk)
            pbar.update(len(chunk))


def load_state_dict_from_url(
    model_key: str,
    entry: WeightEntry,
    *,
    cache_dir: Path | None = None,
    force: bool = False,
) -> dict:
    cache = cache_dir or CACHE_HOME
    cache.mkdir(parents=True, exist_ok=True)
    dst = _cache_path(model_key, entry)

    if not OFFLINE and (force or not dst.exists()):
        tmp_path = Path(tempfile.mkstemp(prefix="lucid-", suffix=".tmp")[1])
        last_err = None
        try:
            for u in _candidate_urls(entry):
                try:
                    req = urllib.request.Request(u)
                    if HF_TOKEN and (HF_ENDPOINT in u or "huggingface.co" in u):
                        req.add_header("Authorization", f"Bearer {HF_TOKEN}")
                    with urllib.request.urlopen(req, context=_SSL_CONTEXT) as r, open(
                        tmp_path, "wb"
                    ) as out:
                        url_name = Path(urlparse(u).path).name or entry.tag
                        _download_with_progress(
                            r,
                            out,
                            desc=f"Downloading {model_key}:{entry.tag} ({url_name})",
                        )
                    break

                except Exception as e:
                    last_err = e
                    continue
            else:
                if last_err:
                    if isinstance(last_err, urllib.error.URLError) and isinstance(
                        getattr(last_err, "reason", None),
                        ssl.SSLCertVerificationError,
                    ):
                        raise RuntimeError(
                            "SSL certificate verification failed while downloading "
                            "weights. Set LUCID_SSL_CA_BUNDLE (or SSL_CERT_FILE) to "
                            "a valid CA bundle, install certifi, or as a last resort "
                            "set LUCID_SSL_NO_VERIFY=1."
                        ) from last_err
                    raise last_err

            if _hash_file(tmp_path) != entry.sha256:
                raise RuntimeError("Checksum mismatch for downloaded weights.")
            shutil.move(tmp_path, dst)

        finally:
            try:
                tmp_path.unlink(missing_ok=True)
            except Exception:
                pass

    elif not dst.exists():
        raise FileNotFoundError(
            "Weights not present in cache and offline mode is enabled."
        )

    if _hash_file(dst) != entry.sha256:
        dst.unlink()
        return load_state_dict_from_url(model_key, entry, cache_dir=cache, force=True)

    try:
        from safetensors.numpy import load_file
    except Exception as e:
        raise ImportError(
            "safetensors is required to save .safetensors files. "
            "Install with `pip install safetensors`."
        ) from e

    state = load_file(str(dst))
    return OrderedDict((k, v) for k, v in state.items())


def apply(model: nn.Module, weights: Enum | str, *, strict: bool = True) -> Path:
    key = getattr(model, "_alt_name", type(model).__name__).lower()
    entry = resolve(weights, model_key=key)
    state = load_state_dict_from_url(key, entry)

    device = getattr(model, "device", "cpu")
    if device == "gpu":
        cooked = {}
        for k, v in state.items():
            if hasattr(v, "shape"):
                cooked[k] = Tensor(v, device="gpu")
            else:
                cooked[k] = v
        state = cooked

    try:
        model.load_state_dict(
            state,
            strict=strict,
            verbose=True,
            progress_desc=f"Applying {key}:{entry.tag}",
        )
    except TypeError:
        model.load_state_dict(state, strict=strict)
    return _cache_path(key, entry)


def list_available(model_key: str) -> dict[str, WeightEntry]:
    key = _canon(model_key)
    if key not in _REGISTRY:
        return {}
    enum = get_enum(key)
    return {m.name: m.value for m in enum}


__all__ = [
    "WeightEntry",
    "get_enum",
    "resolve",
    "apply",
    "list_available",
    "load_state_dict_from_url",
    *_created_names,
]
